# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: federated_language/proto/data_type.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'federated_language/proto/data_type.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(federated_language/proto/data_type.proto\x12\x12\x66\x65\x64\x65rated_language*\xa6\x02\n\x08\x44\x61taType\x12\x0e\n\nDT_INVALID\x10\x00\x12\x0b\n\x07\x44T_BOOL\x10\n\x12\x0b\n\x07\x44T_INT8\x10\x06\x12\x0c\n\x08\x44T_INT16\x10\x05\x12\x0c\n\x08\x44T_INT32\x10\x03\x12\x0c\n\x08\x44T_INT64\x10\t\x12\x0c\n\x08\x44T_UINT8\x10\x04\x12\r\n\tDT_UINT16\x10\x11\x12\r\n\tDT_UINT32\x10\x16\x12\r\n\tDT_UINT64\x10\x17\x12\x0b\n\x07\x44T_HALF\x10\x13\x12\x0c\n\x08\x44T_FLOAT\x10\x01\x12\r\n\tDT_DOUBLE\x10\x02\x12\x10\n\x0c\x44T_COMPLEX64\x10\x08\x12\x11\n\rDT_COMPLEX128\x10\x12\x12\x0f\n\x0b\x44T_BFLOAT16\x10\x0e\x12\r\n\tDT_STRING\x10\x07\"\x04\x08\x0b\x10\x0b\"\x04\x08\x0c\x10\x0c\"\x04\x08\r\x10\r\"\x04\x08\x0f\x10\x0f\"\x04\x08\x10\x10\x10\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'federated_language.proto.data_type_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_DATATYPE']._serialized_start=65
  _globals['_DATATYPE']._serialized_end=359
# @@protoc_insertion_point(module_scope)
