# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: federated_language/proto/array.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'federated_language/proto/array.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from federated_language.proto import data_type_pb2 as federated__language_dot_proto_dot_data__type__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$federated_language/proto/array.proto\x12\x12\x66\x65\x64\x65rated_language\x1a(federated_language/proto/data_type.proto\"/\n\nArrayShape\x12\x0b\n\x03\x64im\x18\x01 \x03(\x03\x12\x14\n\x0cunknown_rank\x18\x02 \x01(\x08\"\xa6\n\n\x05\x41rray\x12+\n\x05\x64type\x18\x01 \x01(\x0e\x32\x1c.federated_language.DataType\x12-\n\x05shape\x18\x02 \x01(\x0b\x32\x1e.federated_language.ArrayShape\x12\x14\n\x07\x63ontent\x18\x12 \x01(\x0cH\x01\x88\x01\x01\x12\x37\n\tbool_list\x18\x03 \x01(\x0b\x32\".federated_language.Array.BoolListH\x00\x12\x36\n\tint8_list\x18\x04 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12\x37\n\nint16_list\x18\x05 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12\x37\n\nint32_list\x18\x06 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12\x39\n\nint64_list\x18\x07 \x01(\x0b\x32#.federated_language.Array.Int64ListH\x00\x12\x37\n\nuint8_list\x18\x08 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12\x38\n\x0buint16_list\x18\t \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12;\n\x0buint32_list\x18\n \x01(\x0b\x32$.federated_language.Array.Uint32ListH\x00\x12;\n\x0buint64_list\x18\x0b \x01(\x0b\x32$.federated_language.Array.Uint64ListH\x00\x12\x39\n\x0c\x66loat16_list\x18\x0c \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12;\n\x0c\x66loat32_list\x18\r \x01(\x0b\x32#.federated_language.Array.FloatListH\x00\x12<\n\x0c\x66loat64_list\x18\x0e \x01(\x0b\x32$.federated_language.Array.DoubleListH\x00\x12=\n\x0e\x63omplex64_list\x18\x0f \x01(\x0b\x32#.federated_language.Array.FloatListH\x00\x12?\n\x0f\x63omplex128_list\x18\x10 \x01(\x0b\x32$.federated_language.Array.DoubleListH\x00\x12:\n\rbfloat16_list\x18\x13 \x01(\x0b\x32!.federated_language.Array.IntListH\x00\x12:\n\x0bstring_list\x18\x11 \x01(\x0b\x32#.federated_language.Array.BytesListH\x00\x1a\x19\n\x08\x42oolList\x12\r\n\x05value\x18\x01 \x03(\x08\x1a\x18\n\x07IntList\x12\r\n\x05value\x18\x01 \x03(\x05\x1a\x1a\n\tInt64List\x12\r\n\x05value\x18\x01 \x03(\x03\x1a\x1b\n\nUint32List\x12\r\n\x05value\x18\x01 \x03(\r\x1a\x1b\n\nUint64List\x12\r\n\x05value\x18\x01 \x03(\x04\x1a\x1a\n\tFloatList\x12\r\n\x05value\x18\x01 \x03(\x02\x1a\x1b\n\nDoubleList\x12\r\n\x05value\x18\x01 \x03(\x01\x1a\x1a\n\tBytesList\x12\r\n\x05value\x18\x01 \x03(\x0c\x42\x06\n\x04kindB\n\n\x08_contentb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'federated_language.proto.array_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ARRAYSHAPE']._serialized_start=102
  _globals['_ARRAYSHAPE']._serialized_end=149
  _globals['_ARRAY']._serialized_start=152
  _globals['_ARRAY']._serialized_end=1470
  _globals['_ARRAY_BOOLLIST']._serialized_start=1228
  _globals['_ARRAY_BOOLLIST']._serialized_end=1253
  _globals['_ARRAY_INTLIST']._serialized_start=1255
  _globals['_ARRAY_INTLIST']._serialized_end=1279
  _globals['_ARRAY_INT64LIST']._serialized_start=1281
  _globals['_ARRAY_INT64LIST']._serialized_end=1307
  _globals['_ARRAY_UINT32LIST']._serialized_start=1309
  _globals['_ARRAY_UINT32LIST']._serialized_end=1336
  _globals['_ARRAY_UINT64LIST']._serialized_start=1338
  _globals['_ARRAY_UINT64LIST']._serialized_end=1365
  _globals['_ARRAY_FLOATLIST']._serialized_start=1367
  _globals['_ARRAY_FLOATLIST']._serialized_end=1393
  _globals['_ARRAY_DOUBLELIST']._serialized_start=1395
  _globals['_ARRAY_DOUBLELIST']._serialized_end=1422
  _globals['_ARRAY_BYTESLIST']._serialized_start=1424
  _globals['_ARRAY_BYTESLIST']._serialized_end=1450
# @@protoc_insertion_point(module_scope)
