# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines an abstract interface for things that possess type signatures."""

import abc

from federated_language.types import computation_types


class TypedObject(abc.ABC):
  """An abstract interface for things that possess type signatures."""

  @property
  @abc.abstractmethod
  def type_signature(self) -> computation_types.Type:
    """Returns the instance of `federated_language.Type` of this object."""
    raise NotImplementedError
